﻿' ---------------------------------------------------------------------
'                     Access エクスポートツール
' ---------------------------------------------------------------------
'
'   このプログラムは、指定されたディレクトリ以下にあるAccessのMDB/ACCDB
'   ファイルを探し、Accessファイルに含まれるテーブル、クエリ、フォーム、
'   レポート、マクロに含まれる各要素(フィールド定義やコントロールなど)の
'   情報をExcelシートに一覧出力するツールです。
'
'   Accessファイルに含まれるモジュールについては、指定されたディレクトリ
'   以下にテキストファイルとしてエクスポートされます。
'
'   動作確認はAccess 2013(Office Professional 2013)で行っています。
'   それ以外のAccessをお使いの場合、お使いのAccessのバージョンにあわせた
'   COMオブジェクトを参照設定で設定しなおし、リコンパイルすることでで動作
'   する可能性があります。
'
'   Ver 1.0     2020/02/02  初版公開
'   Ver 1.0.1   2020/02/03  SendInputのあとにSleepを追加し時間調整を行う
'                           ように変更
'   Ver 1.0.2   2020/02/03  シートに出力する文字列をチェックし、パター
'                           ンにマッチするとき強調表示するように修正
'   Ver 1.0.3   2020/02/05  クエリのSQLを出力するように修正
'                           強調表示するパターンを見つけた場合、マー
'                           カーを出力するように修正
'                           (Excelのデータフィルタで絞り込みしやすくする)
'
'
'   最新版は http://www.ria-lab.com/ からダウンロードしてください。
'
' ---------------------------------------------------------------------
'
' This software is distributed under the license of NYSL.
' ( http://www.kmonos.net/nysl/ )
'
'   NYSL Version 0.9982
'
'	A. 本ソフトウェアは Everyone'sWare です。このソフトを手にした一人一人が、
'	   ご自分の作ったものを扱うのと同じように、自由に利用することが出来ます。
'	
'	  A-1. フリーウェアです。作者からは使用料等を要求しません。
'	  A-2. 有料無料や媒体の如何を問わず、自由に転載・再配布できます。
'	  A-3. いかなる種類の 改変・他プログラムでの利用 を行っても構いません。
'	  A-4. 変更したものや部分的に使用したものは、あなたのものになります。
'	       公開する場合は、あなたの名前の下で行って下さい。
'	
'	B. このソフトを利用することによって生じた損害等について、作者は
'	   責任を負わないものとします。各自の責任においてご利用下さい。
'	
'	C. 著作者人格権は 白府 誠 に帰属します。著作権は放棄します。
'	
'	D. 以上の３項は、ソース・実行バイナリの双方に適用されます。'   
'
'
Imports System.IO
Imports Microsoft.Office.Interop
Imports Microsoft.Office.Interop.Access
Imports Microsoft.Vbe.Interop
Imports NPOI.XSSF.UserModel
Imports Microsoft.Office.Interop.Access.Dao
Imports DataTypeEnum = Microsoft.Office.Interop.Access.Dao.DataTypeEnum
Imports System.Text
Imports System.Runtime.InteropServices
Imports System.ComponentModel

Public Class Form1

#Region "API関連の定義"

    Private Const VK_SHIFT As Short = &H10
    Private Const KEYEVENTF_KEYDOWN As Integer = &H0
    Private Const KEYEVENTF_EXTENDEDKEY As Integer = &H1
    Private Const KEYEVENTF_KEYUP As Integer = &H2
    Private Const INPUT_KEYBOARD As Integer = 1

    <StructLayout(LayoutKind.Sequential, CharSet:=CharSet.Auto)>
    Private Structure KEYBDINPUT
        Public wVk As Short
        Public wScan As Short
        Public dwFlags As Integer
        Public time As Integer
        Public dwExtraInfo As IntPtr
        Public dumy1 As Integer
        Public dumy2 As Integer
    End Structure

    <StructLayout(LayoutKind.Sequential, Size:=28)>
    Private Structure INPUT_TYPE
        Public wType As Integer
        Public ki As KEYBDINPUT
    End Structure

    <DllImport("user32.dll", CharSet:=CharSet.Auto)>
    Private Shared Function SendInput(ByVal nInputs As Integer, ByVal pInputs() As INPUT_TYPE, ByVal cbsize As Integer) As Integer
    End Function

    <DllImport("user32.dll", CharSet:=CharSet.Auto)>
    Private Shared Function MapVirtualKey(ByVal wCode As Integer, ByVal wMapType As Integer) As Integer
    End Function

    <DllImport("user32.dll", CharSet:=CharSet.Auto)>
    Private Shared Function GetMessageExtraInfo() As IntPtr
    End Function


    <DllImport("kernel32.dll", CharSet:=CharSet.Auto)>
    Private Shared Sub Sleep(ByVal dwMilliseconds As Integer)
    End Sub


#End Region

    ''' <summary>
    ''' レポート作成時の情報
    ''' </summary>
    Private Class ExportSheetInfo
        Public Sheet As XSSFSheet
        Public RowNo As Integer
        Public Count As Integer
    End Class

    ''' <summary>
    ''' レポート出力時の出力先カラム番号
    ''' </summary>
    Private Enum REPORT_COLUMNS As Integer
        NO = 0
        OBJECT_TYPE = 1
        OBJECT_SUBTYPE = 2
        OBJECT_PARENT = 3
        OBJECT_NAME = 4
        FIELD_NAME = 5
        FIELD_TYPE = 6
        DATA_SIZE = 7
        CONNECT = 8
        DESCRIPTION = 9
        REMARKS = 10
        MATCHED_MARKER = 11
    End Enum

    ''' <summary>レポート出力時のデータ開始行番号</summary>
    Private Const FIRST_DATA_ROW_NO = 3

    ''' <summary>ファイル名として使用できない文字</summary>
    Private Shared ReadOnly _FILE_NAME_RE As New RegularExpressions.Regex("[\\/:\*\?"" <>\|]")
    ''' <summary>コントロールソースが取得可能なコントロール名</summary>
    Private Shared ReadOnly CONTROLSOURCE_CONTROLS = New HashSet(Of String)(
        {"BoundObjectFrameClass", "ListBoxClass", "CheckBoxClass", "OptionButtonClass",
        "ComboBoxClass", "OptionGroupClass", "CustomControlClass", "TextBoxClass",
        "GroupLevelClass", "ToggleButtonClass"})

    ''' <summary>フォーム/レポートのセクションインデックス</summary>
    Private Shared ReadOnly SECTION_INDEXES = {AcSection.acDetail, AcSection.acHeader,
        AcSection.acFooter, AcSection.acPageHeader, AcSection.acPageFooter,
        AcSection.acGroupLevel1Header, AcSection.acGroupLevel1Footer,
        AcSection.acGroupLevel2Header, AcSection.acGroupLevel2Footer}


    ''' <summary>Accessアプリケーション</summary>
    Private _AC_APP As Access.Application
    ''' <summary>レポート出力用のブック(NPOI)</summary>
    Private _SUMMARY_REPORT As XSSFWorkbook

    ''' <summary>強調表示のパターン</summary>
    Private _HIGHLIGHTING_PATTERN As RegularExpressions.Regex

#Region "イベントハンドラ"

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub Form1_Load(sender As Object, e As EventArgs) Handles MyBase.Load

        TxtSourceDirectory.Text = ""
        TxtDestinationDirectory.Text = ""

    End Sub

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub Form1_Closing(sender As Object, e As CancelEventArgs) Handles Me.Closing

        Dim result = MessageBox.Show("終了しますか?", "確認", MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2)
        If result <> DialogResult.Yes Then
            e.Cancel = True
            Return
        End If

    End Sub

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub BtnSourceDirectory_Click(sender As Object, e As EventArgs) Handles BtnSourceDirectory.Click

        If String.IsNullOrEmpty(TxtSourceDirectory.Text) Then
            SourceFolderBrowserDialog.RootFolder = Environment.SpecialFolder.MyComputer
        Else
            SourceFolderBrowserDialog.SelectedPath = TxtSourceDirectory.Text
        End If

        Dim result = SourceFolderBrowserDialog.ShowDialog()
        If result <> DialogResult.OK Then
            Return
        End If

        TxtSourceDirectory.Text = SourceFolderBrowserDialog.SelectedPath

    End Sub

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub BtnDestinationDirectory_Click(sender As Object, e As EventArgs) Handles BtnDestinationDirectory.Click


        If String.IsNullOrEmpty(TxtDestinationDirectory.Text) Then
            DestinationFolderBrowserDialog.RootFolder = Environment.SpecialFolder.MyComputer
        Else
            DestinationFolderBrowserDialog.SelectedPath = TxtDestinationDirectory.Text
        End If

        Dim result = DestinationFolderBrowserDialog.ShowDialog()
        If result <> DialogResult.OK Then
            Return
        End If

        TxtDestinationDirectory.Text = DestinationFolderBrowserDialog.SelectedPath
    End Sub

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub ChkOnlyVbaExport_CheckedChanged(sender As Object, e As EventArgs) Handles ChkOnlyVbaExport.CheckedChanged
        ChkIgnoreDescription.Enabled = Not ChkOnlyVbaExport.Checked
    End Sub

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub BtnPerform_Click(sender As Object, e As EventArgs) Handles BtnPerform.Click

        If String.IsNullOrEmpty(TxtSourceDirectory.Text.Trim) _
            OrElse String.IsNullOrEmpty(TxtSourceDirectory.Text.Trim) Then
            MessageBox.Show("ディレクトリを指定してください", "警告", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
            Return
        End If

        If Not Directory.Exists(TxtSourceDirectory.Text.Trim) Then
            MessageBox.Show("データベースの格納ディレクトリが存在しません", "警告", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
            Return
        End If
        If Not Directory.Exists(TxtDestinationDirectory.Text.Trim) Then
            MessageBox.Show("エクスポート先ディレクトリが存在しません", "警告", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
            Return
        End If

        Dim pattern = TxtHighlightingPattern.Text.Trim
        If String.IsNullOrEmpty(pattern) Then
            _HIGHLIGHTING_PATTERN = Nothing
        Else
            Try
                _HIGHLIGHTING_PATTERN = New RegularExpressions.Regex(pattern)
            Catch ex As Exception
                MessageBox.Show("正規表現に誤りがあります", "警告", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
                Return
            End Try
        End If

        Dim result = MessageBox.Show("エクスポート処理を実行しますか?", "確認", MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2)
        If result <> DialogResult.Yes Then
            Return
        End If

        Perform(TxtSourceDirectory.Text.Trim, TxtDestinationDirectory.Text.Trim)

    End Sub

#End Region


#Region "処理"

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="sourceDirectory"></param>
    ''' <param name="destinationDirectory"></param>
    Private Sub Perform(sourceDirectory As String, destinationDirectory As String)

        BtnSourceDirectory.Enabled = False
        BtnDestinationDirectory.Enabled = False
        BtnPerform.Enabled = False
        TxtHighlightingPattern.Enabled = False
        ChkIgnoreDescription.Enabled = False
        ChkOnlyVbaExport.Enabled = False

        TxtCurrentDatabase.Text = ""
        TxtCurrentCategory.Text = ""
        TxtCurrentObject.Text = ""

        Dim result As Boolean
        Dim message As String

        Try

            _AC_APP = New Access.Application
            _AC_APP.Visible = False

            ' 出力先のブックを作成
            _SUMMARY_REPORT = NPOI.SS.UserModel.WorkbookFactory.Create("resources\TEMPLATE.xlsx")

            ' 指定ディレクトリからAccess DBを探しながらエクスポートする
            FindAccessDb(sourceDirectory, destinationDirectory)

            ' テンプレートシートを非表示に
            _SUMMARY_REPORT.SetSheetHidden(0, NPOI.SS.UserModel.SheetState.VeryHidden)

            ' エクスポート先にブックを出力
            Dim summaryName = "SUMMARY_" & DateTime.Now.ToString("yyyyMMdd_HHmmss") & ".xlsx"
            Dim reportPath = Path.Combine(destinationDirectory, summaryName)
            Using stream = New FileStream(reportPath, FileMode.Create)
                _SUMMARY_REPORT.Write(stream)
            End Using

            TxtCurrentDatabase.Text = ""
            TxtCurrentCategory.Text = ""
            TxtCurrentObject.Text = ""

            result = True
            message = ""

        Catch ex As Exception

            result = False
            message = ex.Message

        Finally

            _AC_APP.Quit()
            Marshal.ReleaseComObject(_AC_APP)
            _AC_APP = Nothing

            Me.Enabled = True

            BtnSourceDirectory.Enabled = True
            BtnDestinationDirectory.Enabled = True
            BtnPerform.Enabled = True
            TxtHighlightingPattern.Enabled = True
            ChkIgnoreDescription.Enabled = Not ChkOnlyVbaExport.Checked
            ChkOnlyVbaExport.Enabled = False

        End Try

        If result Then
            MessageBox.Show("処理が完了しました", "完了", MessageBoxButtons.OK, MessageBoxIcon.Information)
        Else
            MessageBox.Show("処理中にエラーが発生しました。" & vbCrLf &
                            "(" & message & ")",
                            "エラー", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End If

    End Sub

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="sourceDirectory"></param>
    ''' <param name="destinationDirectory"></param>
    Private Sub FindAccessDb(sourceDirectory As String, destinationDirectory As String)

        Dim di = New DirectoryInfo(sourceDirectory)
        For Each fi In di.GetFiles("*")

            ' 拡張子がACCDBまたはMDBのものを処理対象とする
            ' 対象とするファイル種別を増やしたいときはここを修正する
            Dim Extension = Path.GetExtension(fi.Name).ToLower
            If Extension <> ".accdb" AndAlso Extension <> ".mdb" AndAlso Extension <> ".mda" Then
                Continue For
            End If

            TxtCurrentDatabase.Text = fi.FullName
            ExportAccessDb(fi.FullName, destinationDirectory)

        Next

        For Each subdi In di.GetDirectories()
            FindAccessDb(subdi.FullName, destinationDirectory)
        Next

    End Sub


    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="dbPath"></param>
    ''' <param name="exportRootPath"></param>
    Private Sub ExportAccessDb(dbPath As String, exportRootPath As String)

        Console.WriteLine(dbPath)

        '--
        ' AutoExecマクロの実行を抑制するためSHIFTキーを押下した状態で開くことを再現
        '--

        Dim KBDIN(0) As INPUT_TYPE
        ' SHIFTキーダウン
        With KBDIN(0)
            .wType = INPUT_KEYBOARD
            .ki.wVk = VK_SHIFT
            .ki.wScan = MapVirtualKey(VK_SHIFT, 0)
            .ki.dwFlags = KEYEVENTF_KEYDOWN Or KEYEVENTF_EXTENDEDKEY
            .ki.time = 0
            .ki.dwExtraInfo = GetMessageExtraInfo()
        End With
        SendInput(1, KBDIN, Marshal.SizeOf(GetType(INPUT_TYPE)))

        Sleep(100)

        ' Accessのファイルを開く
        _AC_APP.OpenCurrentDatabase(dbPath, Exclusive:=True)

        ' SHIFTキーアップ
        KBDIN(0).ki.dwFlags = KEYEVENTF_KEYUP Or KEYEVENTF_EXTENDEDKEY
        SendInput(1, KBDIN, Marshal.SizeOf(GetType(INPUT_TYPE)))

        Sleep(100)

        Try

            '--
            ' MDBに含まれるモジュールのエクスポート
            '--
            TxtCurrentCategory.Text = "VBAモジュールのエクスポート"

            Dim exportDirectory = Path.Combine(exportRootPath,
                                           GetPathWithoutDriveName(Directory.GetParent(dbPath).FullName),
                                           Path.GetFileNameWithoutExtension(dbPath) & "_" & Path.GetExtension(dbPath))

            If Not Directory.Exists(exportDirectory) Then
                Directory.CreateDirectory(exportDirectory)
            End If
            ExportAllVbaModules(exportDirectory)

            ' VBA出力のみのときはここで処理を打ち切り
            If ChkOnlyVbaExport.Checked Then
                Return
            End If

            Dim ESI As New ExportSheetInfo
            With ESI
                .Sheet = _SUMMARY_REPORT.CloneSheet(0)
                .RowNo = FIRST_DATA_ROW_NO
                .Count = 0
            End With

            '--
            ' Excelシート名とMDBファイル名の設定
            '--
            SetSheetName(ESI.Sheet, Path.GetFileNameWithoutExtension(dbPath))
            GetCell(GetRow(ESI.Sheet, 0), 2).SetCellValue(dbPath)

            '--
            ' MDBに含まれる各要素をExcelシートに出力
            '--

            TxtCurrentCategory.Text = "レポート作成 - テーブル"
            ExportTables(ESI)
            TxtCurrentCategory.Text = "レポート作成 - クエリ"
            ExportQueries(ESI)
            TxtCurrentCategory.Text = "レポート作成 - フォーム"
            ExportForms(ESI)
            TxtCurrentCategory.Text = "レポート作成 - レポート"
            ExportReports(ESI)
            TxtCurrentCategory.Text = "レポート作成 - モジュール"
            ExportModules(ESI)
            TxtCurrentCategory.Text = "レポート作成 - マクロ"
            ExportMacros(ESI)

            ' スタイルを適用
            SetSheetStyle(ESI.Sheet)
            ' 強調表示対象のチェック
            If _HIGHLIGHTING_PATTERN IsNot Nothing Then
                CheckHighlighting(ESI.Sheet)
            End If

            Console.WriteLine(vbTab & "Count=" & ESI.Count)

        Catch ex As Exception
            Console.WriteLine(ex.Message)
            Console.WriteLine(ex.StackTrace)
        Finally
            Try
                _AC_APP.CloseCurrentDatabase()
            Catch ex2 As Exception
                Console.WriteLine(ex2.StackTrace)
            End Try
        End Try

    End Sub


#Region "エクスポート"

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="ESI"></param>
    Private Sub ExportTables(ESI As ExportSheetInfo)

        Console.WriteLine("【テーブル】")

        Dim tbldefs As Access.Dao.TableDefs = _AC_APP.CurrentDb.TableDefs
        For i = 0 To tbldefs.Count - 1

            Try

                Dim tbldef As Access.Dao.TableDef = tbldefs.Item(i)
                If tbldef.Name.StartsWith("MSys") Then
                    Continue For
                End If

                TxtCurrentObject.Text = tbldef.Name
                Console.WriteLine(vbTab & "+- " & tbldef.Name)

                Dim row = GetRow(ESI.Sheet, ESI.RowNo)
                GetCell(row, REPORT_COLUMNS.NO).SetCellValue(ESI.Count + 1)
                GetCell(row, REPORT_COLUMNS.OBJECT_TYPE).SetCellValue("テーブル")
                GetCell(row, REPORT_COLUMNS.OBJECT_SUBTYPE).SetCellValue(If(String.IsNullOrEmpty(tbldef.Connect), "", "リンク"))
                GetCell(row, REPORT_COLUMNS.OBJECT_NAME).SetCellValue(tbldef.Name)
                GetCell(row, REPORT_COLUMNS.CONNECT).SetCellValue(tbldef.Connect)
                GetCell(row, REPORT_COLUMNS.DESCRIPTION).SetCellValue(If(ChkIgnoreDescription.Checked, "", GetDescription(tbldef)))

                ESI.RowNo += 1
                ESI.Count += 1

                For j = 0 To tbldef.Fields.Count - 1

                    Dim field As Access.Dao.Field = tbldef.Fields.Item(j)

                    Dim row2 = GetRow(ESI.Sheet, ESI.RowNo)
                    GetCell(row2, REPORT_COLUMNS.NO).SetCellValue(ESI.Count + 1)
                    GetCell(row2, REPORT_COLUMNS.OBJECT_TYPE).SetCellValue("")
                    GetCell(row2, REPORT_COLUMNS.OBJECT_SUBTYPE).SetCellValue("")
                    With GetCell(row2, REPORT_COLUMNS.FIELD_NAME)
                        .SetCellValue(field.Name)
                    End With
                    GetCell(row2, REPORT_COLUMNS.FIELD_TYPE).SetCellValue(GetFieldTypeName(field.Type))
                    GetCell(row2, REPORT_COLUMNS.DATA_SIZE).SetCellValue(field.Size)
                    GetCell(row2, REPORT_COLUMNS.DESCRIPTION).SetCellValue(If(ChkIgnoreDescription.Checked, "", GetDescription(field)))

                    Marshal.ReleaseComObject(field)
                    field = Nothing

                    ESI.RowNo += 1
                    ESI.Count += 1

                Next

                _AC_APP.DoCmd.Close(AcObjectType.acTable, tbldef.Name, AcCloseSave.acSaveNo)

                Marshal.ReleaseComObject(tbldef)
                tbldef = Nothing

            Catch ex As Exception

                Dim str = ex.StackTrace
                Dim row = GetRow(ESI.Sheet, ESI.RowNo)
                GetCell(row, REPORT_COLUMNS.NO).SetCellValue(ESI.Count + 1)
                GetCell(row, REPORT_COLUMNS.REMARKS).SetCellValue(If(str.Length <= 32767, str, str.Substring(0, 32767)))

                ESI.RowNo += 1
                ESI.Count += 1

            End Try

            System.Windows.Forms.Application.DoEvents()

        Next

        Marshal.ReleaseComObject(tbldefs)
        tbldefs = Nothing

    End Sub

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="ESI"></param>
    Private Sub ExportQueries(ESI As ExportSheetInfo)

        Console.WriteLine("【クエリ】")

        Dim qrydefs As Access.Dao.QueryDefs = _AC_APP.CurrentDb.QueryDefs
        For i = 0 To qrydefs.Count - 1

            Try

                Dim qrydef As Access.Dao.QueryDef = qrydefs.Item(i)

                TxtCurrentObject.Text = qrydef.Name
                Console.WriteLine(vbTab & "+- " & qrydef.Name)

                Dim row = GetRow(ESI.Sheet, ESI.RowNo)
                GetCell(row, REPORT_COLUMNS.NO).SetCellValue(ESI.Count + 1)
                GetCell(row, REPORT_COLUMNS.OBJECT_TYPE).SetCellValue("クエリ")
                GetCell(row, REPORT_COLUMNS.OBJECT_SUBTYPE).SetCellValue(If(String.IsNullOrEmpty(qrydef.Connect), "", "リンク"))
                GetCell(row, REPORT_COLUMNS.OBJECT_NAME).SetCellValue(qrydef.Name)
                GetCell(row, REPORT_COLUMNS.FIELD_NAME).SetCellValue(If(qrydef.SQL.Length <= 32767, qrydef.SQL, qrydef.SQL.Substring(0, 32767)))
                GetCell(row, REPORT_COLUMNS.DESCRIPTION).SetCellValue(If(ChkIgnoreDescription.Checked, "", GetDescription(qrydef)))
                GetCell(row, REPORT_COLUMNS.REMARKS).SetCellValue(If(qrydef.SQL.Length > 32767, "※継続するSQLあり", ""))

                ESI.RowNo += 1
                ESI.Count += 1

                For j = 0 To qrydef.Fields.Count - 1

                    Dim field As Access.Dao.Field = qrydef.Fields.Item(j)

                    Dim row2 = GetRow(ESI.Sheet, ESI.RowNo)
                    GetCell(row2, REPORT_COLUMNS.NO).SetCellValue(ESI.Count + 1)
                    GetCell(row2, REPORT_COLUMNS.OBJECT_TYPE).SetCellValue("")
                    GetCell(row2, REPORT_COLUMNS.OBJECT_SUBTYPE).SetCellValue("")
                    With GetCell(row2, REPORT_COLUMNS.FIELD_NAME)
                        .SetCellValue(field.Name)
                    End With
                    GetCell(row2, REPORT_COLUMNS.FIELD_TYPE).SetCellValue(GetFieldTypeName(field.Type))
                    GetCell(row2, REPORT_COLUMNS.DATA_SIZE).SetCellValue(field.Size)
                    GetCell(row2, REPORT_COLUMNS.DESCRIPTION).SetCellValue(If(ChkIgnoreDescription.Checked, "", GetDescription(field)))

                    Marshal.ReleaseComObject(field)
                    field = Nothing

                    ESI.RowNo += 1
                    ESI.Count += 1

                Next

                _AC_APP.DoCmd.Close(AcObjectType.acQuery, qrydef.Name, AcCloseSave.acSaveNo)
                qrydef.Close()

                Marshal.ReleaseComObject(qrydef)
                qrydef = Nothing

            Catch ex As Exception

                Dim str = ex.StackTrace
                Dim row = GetRow(ESI.Sheet, ESI.RowNo)
                GetCell(row, REPORT_COLUMNS.NO).SetCellValue(ESI.Count + 1)
                GetCell(row, REPORT_COLUMNS.REMARKS).SetCellValue(If(str.Length <= 32767, str, str.Substring(0, 32767)))

                ESI.RowNo += 1
                ESI.Count += 1

            End Try

            System.Windows.Forms.Application.DoEvents()

        Next

        Marshal.ReleaseComObject(qrydefs)
        qrydefs = Nothing

    End Sub

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="ESI"></param>
    Private Sub ExportForms(ESI As ExportSheetInfo)

        Console.WriteLine("【フォーム】")

        Dim frms As AllForms = _AC_APP.CurrentProject.AllForms
        For i = 0 To frms.Count - 1

            Dim obj As Access.AccessObject = frms.Item(i)

            TxtCurrentObject.Text = obj.Name
            Console.WriteLine(vbTab & "+- " & obj.Name)

            _AC_APP.DoCmd.OpenForm(obj.Name, AcView.acViewDesign)

            Try

                Dim frm As Access.Form = DirectCast(_AC_APP.Forms.Item(obj.Name), Access.Form)

                Dim row = GetRow(ESI.Sheet, ESI.RowNo)
                GetCell(row, REPORT_COLUMNS.NO).SetCellValue(ESI.Count + 1)
                GetCell(row, REPORT_COLUMNS.OBJECT_TYPE).SetCellValue("フォーム")
                GetCell(row, REPORT_COLUMNS.OBJECT_NAME).SetCellValue(obj.Name)
                GetCell(row, REPORT_COLUMNS.FIELD_NAME).SetCellValue(frm.RecordSource)
                GetCell(row, REPORT_COLUMNS.DESCRIPTION).SetCellValue(If(ChkIgnoreDescription.Checked, "", GetDescription(obj)))

                ESI.RowNo += 1
                ESI.Count += 1

                For Each sectionIndex In SECTION_INDEXES

                    Try

                        Dim section As Access.Section = frm.Section(sectionIndex)

                        Console.WriteLine(vbTab & vbTab & "+- " & section.Name)

                        Dim ctrls As Access.Children = section.Controls

                        For j = 0 To ctrls.Count - 1

                            Dim ctrl = ctrls.Item(j)
                            Console.WriteLine(vbTab & vbTab & vbTab & "+-> " & ctrl.name)

                            Dim row2 = GetRow(ESI.Sheet, ESI.RowNo)
                            GetCell(row2, REPORT_COLUMNS.NO).SetCellValue(ESI.Count + 1)
                            GetCell(row2, REPORT_COLUMNS.OBJECT_SUBTYPE).SetCellValue(TypeName(ctrl))
                            GetCell(row2, REPORT_COLUMNS.OBJECT_NAME).SetCellValue(ctrl.Name)
                            GetCell(row2, REPORT_COLUMNS.FIELD_NAME).SetCellValue(GetControlSource(ctrl))
                            GetCell(row2, REPORT_COLUMNS.DESCRIPTION).SetCellValue(If(ChkIgnoreDescription.Checked, "", GetDescription(ctrl)))
                            GetCell(row2, REPORT_COLUMNS.OBJECT_PARENT).SetCellValue(section.Name)

                            Marshal.ReleaseComObject(ctrl)
                            ctrl = Nothing

                            ESI.RowNo += 1
                            ESI.Count += 1
                        Next

                        Marshal.ReleaseComObject(ctrls)
                        ctrls = Nothing
                        Marshal.ReleaseComObject(section)
                        section = Nothing

                    Catch ex As COMException
                        If ex.ErrorCode = &H800A099E Then
                            ' セクション番号の指定が正しくありません。
                            Continue For
                        End If
                        Throw ex
                    End Try

                Next

                Console.WriteLine("")

                Marshal.ReleaseComObject(frm)
                frm = Nothing

            Finally

                _AC_APP.DoCmd.Close(AcObjectType.acForm, obj.Name, AcCloseSave.acSaveNo)
                Marshal.ReleaseComObject(obj)
                obj = Nothing

            End Try

        Next

        Marshal.ReleaseComObject(frms)
        frms = Nothing

    End Sub

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="ESI"></param>
    Private Sub ExportReports(ESI As ExportSheetInfo)

        Console.WriteLine("【レポート】")

        Dim rpts As AllReports = _AC_APP.CurrentProject.AllReports
        For i = 0 To rpts.Count - 1

            Dim obj As Access.AccessObject = rpts.Item(i)

            TxtCurrentObject.Text = obj.Name
            Console.WriteLine(vbTab & "+- " & obj.Name)

            _AC_APP.DoCmd.OpenReport(obj.Name, AcView.acViewDesign)

            Try

                Dim rpt As Access.Report = DirectCast(_AC_APP.Reports.Item(obj.Name), Access.Report)

                Dim row = GetRow(ESI.Sheet, ESI.RowNo)
                GetCell(row, REPORT_COLUMNS.NO).SetCellValue(ESI.Count + 1)
                GetCell(row, REPORT_COLUMNS.OBJECT_TYPE).SetCellValue("レポート")
                GetCell(row, REPORT_COLUMNS.OBJECT_NAME).SetCellValue(obj.Name)
                GetCell(row, REPORT_COLUMNS.FIELD_NAME).SetCellValue(rpt.RecordSource)
                GetCell(row, REPORT_COLUMNS.DESCRIPTION).SetCellValue(If(ChkIgnoreDescription.Checked, "", GetDescription(obj)))

                ESI.RowNo += 1
                ESI.Count += 1

                For Each sectionIndex In SECTION_INDEXES

                    Try

                        Dim section As Access.Section = rpt.Section(sectionIndex)

                        Console.WriteLine(vbTab & vbTab & "+- " & section.Name)

                        Dim ctrls As Access.Children = section.Controls

                        For j = 0 To ctrls.Count - 1

                            Dim ctrl = ctrls.Item(j)
                            Console.WriteLine(vbTab & vbTab & vbTab & "+-> " & ctrl.name)

                            Dim row2 = GetRow(ESI.Sheet, ESI.RowNo)
                            GetCell(row2, REPORT_COLUMNS.NO).SetCellValue(ESI.Count + 1)
                            GetCell(row2, REPORT_COLUMNS.OBJECT_SUBTYPE).SetCellValue(TypeName(ctrl))
                            GetCell(row2, REPORT_COLUMNS.OBJECT_NAME).SetCellValue(ctrl.Name)
                            GetCell(row2, REPORT_COLUMNS.FIELD_NAME).SetCellValue(GetControlSource(ctrl))
                            GetCell(row2, REPORT_COLUMNS.DESCRIPTION).SetCellValue(If(ChkIgnoreDescription.Checked, "", GetDescription(ctrl)))
                            GetCell(row2, REPORT_COLUMNS.OBJECT_PARENT).SetCellValue(section.Name)

                            Marshal.ReleaseComObject(ctrl)
                            ctrl = Nothing

                            ESI.RowNo += 1
                            ESI.Count += 1
                        Next

                        Marshal.ReleaseComObject(ctrls)
                        ctrls = Nothing
                        Marshal.ReleaseComObject(section)
                        section = Nothing

                    Catch ex As COMException
                        If ex.ErrorCode = &H800A099E Then
                            ' セクション番号の指定が正しくありません。
                            Continue For
                        End If
                        Throw ex
                    End Try

                Next

                Console.WriteLine("")

                Marshal.ReleaseComObject(rpt)
                rpt = Nothing

            Finally

                _AC_APP.DoCmd.Close(AcObjectType.acReport, obj.Name, AcCloseSave.acSaveNo)
                Marshal.ReleaseComObject(obj)
                obj = Nothing

            End Try

        Next

        Marshal.ReleaseComObject(rpts)
        rpts = Nothing

    End Sub

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="ESI"></param>
    Private Sub ExportMacros(ESI As ExportSheetInfo)

        Console.WriteLine("【マクロ】")

        Dim macros As AllMacros = _AC_APP.CurrentProject.AllMacros

        For i = 0 To macros.Count - 1

            Dim obj As Access.AccessObject = macros.Item(i)

            TxtCurrentObject.Text = obj.Name
            Console.WriteLine(vbTab & "+- " & obj.Name)

            Dim row = GetRow(ESI.Sheet, ESI.RowNo)
            GetCell(row, REPORT_COLUMNS.NO).SetCellValue(ESI.Count + 1)
            GetCell(row, REPORT_COLUMNS.OBJECT_TYPE).SetCellValue("マクロ")
            GetCell(row, REPORT_COLUMNS.OBJECT_NAME).SetCellValue(obj.Name)
            GetCell(row, REPORT_COLUMNS.DESCRIPTION).SetCellValue(If(ChkIgnoreDescription.Checked, "", GetDescription(obj)))

            ESI.RowNo += 1
            ESI.Count += 1
        Next

        Marshal.ReleaseComObject(macros)
        macros = Nothing

    End Sub

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="ESI"></param>
    Private Sub ExportModules(ESI As ExportSheetInfo)

        Console.WriteLine("【モジュール】")

        Dim projects = _AC_APP.VBE.VBProjects

        For i = 0 To projects.Count - 1

            Dim prj As VBProject = projects(i)
            Console.WriteLine(vbTab & "+- " & prj.Name)

            For j = 1 To prj.VBComponents.Count

                Dim VBC As VBComponent = prj.VBComponents.Item(j)

                TxtCurrentObject.Text = prj.Name & " / " & VBC.Name
                Console.WriteLine(vbTab & vbTab & "+- " & prj.Name)

                Dim row = GetRow(ESI.Sheet, ESI.RowNo)
                GetCell(row, REPORT_COLUMNS.NO).SetCellValue(ESI.Count + 1)
                GetCell(row, REPORT_COLUMNS.OBJECT_TYPE).SetCellValue("モジュール")
                GetCell(row, REPORT_COLUMNS.OBJECT_SUBTYPE).SetCellValue(GetModuleExtension(VBC.Type).ToUpper)
                GetCell(row, REPORT_COLUMNS.OBJECT_NAME).SetCellValue(VBC.Name)
                GetCell(row, REPORT_COLUMNS.OBJECT_PARENT).SetCellValue(prj.Name)
                GetCell(row, REPORT_COLUMNS.DESCRIPTION).SetCellValue(If(ChkIgnoreDescription.Checked, "", GetDescription(VBC)))

                Marshal.ReleaseComObject(VBC)
                VBC = Nothing

                ESI.RowNo += 1
                ESI.Count += 1

                Console.WriteLine("")

            Next

            Marshal.ReleaseComObject(prj)
            prj = Nothing

        Next

        Marshal.ReleaseComObject(projects)
        projects = Nothing

    End Sub

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="exportDirectory"></param>
    Private Sub ExportAllVbaModules(exportDirectory As String)


        Console.WriteLine("【VBAモジュール】")

        Dim projects = _AC_APP.VBE.VBProjects
        For i = 0 To projects.Count - 1

            Dim prj As VBProject = projects(i)

            Console.WriteLine(vbTab & "+- " & prj.FileName)

            For j = 1 To prj.VBComponents.Count

                Dim VBC As VBComponent = prj.VBComponents.Item(j)

                TxtCurrentObject.Text = prj.Name & " / " & VBC.Name
                Console.WriteLine(vbTab & vbTab & "+- " & VBC.Name & vbTab & VBC.Type)

                Dim name = DetermineExportFileName(VBC.Name) & "." & GetModuleExtension(VBC.Type)
                Dim targetPath = Path.Combine(exportDirectory, name)

                VBC.Export(targetPath)

                Marshal.ReleaseComObject(VBC)
                VBC = Nothing

            Next

            Marshal.ReleaseComObject(prj)
            prj = Nothing

            Console.WriteLine("")

        Next

        Marshal.ReleaseComObject(projects)
        projects = Nothing

    End Sub

#End Region

#Region "ユーティリティ"

    ''' <summary>
    ''' フォーム/レポートに貼り付けられたコントロールからコントロールソースを取得する
    ''' </summary>
    ''' <param name="ctrl"></param>
    ''' <returns></returns>
    Private Shared Function GetControlSource(ctrl As Object) As String
        Return If(CONTROLSOURCE_CONTROLS.Contains(TypeName(ctrl)), ctrl.ControlSource, "")
    End Function

    ''' <summary>
    ''' 説明プロパティの値を取得する
    ''' </summary>
    ''' <param name="obj"></param>
    ''' <returns></returns>
    Private Shared Function GetDescription(obj As Object) As String
        On Error Resume Next
        Return obj.Properties.Item("Description").Value
    End Function


    ''' <summary>
    ''' モジュールの種類に応じた拡張子を取得する
    ''' </summary>
    ''' <param name="type"></param>
    ''' <returns></returns>
    Private Shared Function GetModuleExtension(type As Integer) As String

        Select Case type
            Case 1
                Return "bas"
            Case 2
                Return "cls"
            Case 3
                Return "frm"
            Case 11
                Return "dsr"
            Case 100
                Return "bas"
        End Select

        Throw New ArgumentException

    End Function

    ''' <summary>
    ''' Accessの型名を取得する
    ''' </summary>
    ''' <param name="type"></param>
    ''' <returns></returns>
    Private Shared Function GetFieldTypeName(type As Short) As String

        Select Case type
            Case DataTypeEnum.dbBigInt
                Return "BigInt"
            Case DataTypeEnum.dbBinary
                Return "Binary"
            Case DataTypeEnum.dbBoolean
                Return "Boolean"
            Case DataTypeEnum.dbByte
                Return "Byte"
            Case DataTypeEnum.dbChar
                Return "Char"
            Case DataTypeEnum.dbCurrency
                Return "Currency"
            Case DataTypeEnum.dbDate
                Return "Date"
            Case DataTypeEnum.dbDecimal
                Return "Decimal"
            Case DataTypeEnum.dbDouble
                Return "Double"
            Case DataTypeEnum.dbFloat
                Return "Float"
            Case DataTypeEnum.dbGUID
                Return "GUID"
            Case DataTypeEnum.dbInteger
                Return "Integer"
            Case DataTypeEnum.dbLong
                Return "Long"
            Case DataTypeEnum.dbLongBinary
                Return "Long Binary"
            Case DataTypeEnum.dbMemo
                Return "Memo"
            Case DataTypeEnum.dbNumeric
                Return "Numeric"
            Case DataTypeEnum.dbSingle
                Return "Single"
            Case DataTypeEnum.dbText
                Return "Text"
            Case DataTypeEnum.dbTime
                Return "Time"
            Case DataTypeEnum.dbTimeStamp
                Return "Timestamp"
            Case DataTypeEnum.dbVarBinary
                Return "VarBinary"
        End Select

        Return "Unknow Type ('" & CStr(type) & ")"

    End Function

#End Region

#End Region

#Region "NPOI/Excel"

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="sheet"></param>
    ''' <param name="rowNo"></param>
    ''' <returns></returns>
    Private Shared Function GetRow(sheet As XSSFSheet, rowNo As Integer) As XSSFRow
        Return If(sheet.GetRow(rowNo), sheet.CreateRow(rowNo))
    End Function

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="row"></param>
    ''' <param name="columnNo"></param>
    ''' <returns></returns>
    Private Shared Function GetCell(row As XSSFRow, columnNo As Integer) As XSSFCell
        Return If(row.GetCell(columnNo), row.CreateCell(columnNo))
    End Function

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="sheet"></param>
    ''' <param name="sheetName"></param>
    Private Sub SetSheetName(sheet As XSSFSheet, sheetName As String)
        Dim sheetIndex = _SUMMARY_REPORT.GetSheetIndex(sheet)
        _SUMMARY_REPORT.SetSheetName(sheetIndex, sheetName)
    End Sub

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="sheet"></param>
    Private Sub SetSheetStyle(sheet As XSSFSheet)

        Dim style = _SUMMARY_REPORT.CreateCellStyle()

        style.BorderTop = NPOI.SS.UserModel.BorderStyle.Thin
        style.BorderLeft = NPOI.SS.UserModel.BorderStyle.Thin
        style.BorderRight = NPOI.SS.UserModel.BorderStyle.Thin
        style.BorderBottom = NPOI.SS.UserModel.BorderStyle.Thin
        style.BorderDiagonal = NPOI.SS.UserModel.BorderStyle.Thin

        style.WrapText = True
        style.VerticalAlignment = NPOI.SS.UserModel.VerticalAlignment.Top

        For rowNo = FIRST_DATA_ROW_NO To sheet.LastRowNum
            Dim row = sheet.GetRow(rowNo)
            For colNo = 0 To REPORT_COLUMNS.MATCHED_MARKER
                Dim cell = GetCell(row, colNo)
                cell.CellStyle = style
            Next
        Next

    End Sub

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="sheet"></param>
    Private Sub CheckHighlighting(sheet As XSSFSheet)

        For rowNo = FIRST_DATA_ROW_NO To sheet.LastRowNum
            Dim row = sheet.GetRow(rowNo)
            Dim marked = False
            For colNo = 0 To REPORT_COLUMNS.DESCRIPTION
                Dim cell = GetCell(row, colNo)
                If cell.CellType <> NPOI.SS.UserModel.CellType.String Then
                    Continue For
                End If
                If Not _HIGHLIGHTING_PATTERN.IsMatch(cell.StringCellValue()) Then
                    Continue For
                End If

                Dim style = _SUMMARY_REPORT.CreateCellStyle()
                style.CloneStyleFrom(cell.CellStyle)

                Dim font = _SUMMARY_REPORT.CreateFont()
                font.IsBold = True
                font.FontHeightInPoints = cell.CellStyle.GetFont(_SUMMARY_REPORT).FontHeightInPoints
                style.SetFont(font)

                style.FillPattern = NPOI.SS.UserModel.FillPattern.SolidForeground
                style.FillForegroundColor = NPOI.SS.UserModel.IndexedColors.LightYellow.Index

                cell.CellStyle = style

                If marked Then
                    Continue For
                End If

                Dim cellMatched = GetCell(row, REPORT_COLUMNS.MATCHED_MARKER)
                cellMatched.SetCellValue("*")

                Dim styleMatched = _SUMMARY_REPORT.CreateCellStyle()
                styleMatched.CloneStyleFrom(cellMatched.CellStyle)
                styleMatched.Alignment = NPOI.SS.UserModel.HorizontalAlignment.Center

                cellMatched.CellStyle = styleMatched

                marked = True

            Next
        Next

    End Sub

#End Region

#Region "その他処理/ユーティリティ"

    ''' <summary>
    ''' エクスポートファイル名を決定する
    ''' </summary>
    ''' <param name="filename"></param>
    ''' <returns></returns>
    Private Function DetermineExportFileName(filename As String) As String
        Return _FILE_NAME_RE.Replace(filename, "_")
    End Function

    ''' <summary>
    ''' ドライブ名を除いたパスを取得する
    ''' </summary>
    ''' <param name="fullPath"></param>
    ''' <returns></returns>
    Private Shared Function GetPathWithoutDriveName(fullPath As String) As String
        Return fullPath.Substring(Path.GetPathRoot(fullPath).Length)
    End Function


#End Region

End Class
